=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed 24 June 2012 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for commercial purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   APuzzleJimSolo_byFredo6.rb
# Original Date	:   24 June 2012 - version 1.0
# Description	:   Resolution of a Puzzle proposed by Solo and Jim
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module PuzzleJimSolo_byFredo6

def self.startup ; resolve ; end
def self.resolve
	puts "\n======================================="
	puts "Puzzle Solo - Jim - Proposal by Fredo6"
	puts "======================================="

	#Registering the 8 equations: [ [index of variables], proc to test]
	@lst_eq = []
	@lst_eq[0] = [[ 0,  1,  2,  3], proc { |v0, v1, v2, v3| v0 + v1 * v2 - v3 == 27 }]
	@lst_eq[1] = [[ 4,  5,  6,  7], proc { |v0, v1, v2, v3| v0 + v1 * v2 + v3 == 56 }]
	@lst_eq[2] = [[ 8,  9, 10, 11], proc { |v0, v1, v2, v3| v0 * v1 + v2 - v3 == 85 }]
	@lst_eq[3] = [[12, 13, 14, 15], proc { |v0, v1, v2, v3| v0 - v1 + int_div(v2, v3) == -6 }]
	@lst_eq[4] = [[ 0,  4,  8, 12], proc { |v0, v1, v2, v3| v0 + v1 * v2 - v3 == 105 }]
	@lst_eq[5] = [[ 1,  5,  9, 13], proc { |v0, v1, v2, v3| v0 * v1 + v2 + v3 == 129 }]
	@lst_eq[6] = [[ 2,  6, 10, 14], proc { |v0, v1, v2, v3| v0 - v1 - v2 + v3 == 9 }]
	@lst_eq[7] = [[ 3,  7, 11, 15], proc { |v0, v1, v2, v3| v0 + v1 + v2 - v3 == 31 }]
	
	#Resolving the equations in some order
	t0 = Time.now
	@all_val = (1..16).to_a
	@nb_try = 0					#Number of calls to an equation test proc
	@lst_sol = [[]]				#List of solutions
	[3, 4, 2, 5, 1, 7, 0, 6].each do |i|
		resolve_equation @lst_eq[i]
		break if @lst_sol.length < 2
	end	
	puts "time = #{Time.now - t0} seconds - nb trials = #{@nb_try}\nSolution = #{@lst_sol[0].inspect}"
end

#Resolve one equation with variables found from previous equations
def self.resolve_equation(leq)
	lvar, proc = leq		#lvar = index of variables in the 8-variable list
	j0, j1, j2, j3 = lvar	#Index of the 4 variables to be used for equation
	lst_newsol = []			#List of solutions refined with current equation
	
	#Loop on testing the equation with all possible values, taking the already-computed solutions
	@lst_sol.each do |sol_orig|
		vs0, vs1, vs2, vs3 = lvar.collect { |j| sol_orig[j] }	#Variables already set (a value or nil)
		sol = sol_orig.clone									#Placeholder for computing solutions
		val_left = @all_val - sol.compact						#Values possible to make sure all values are different
		((vs0) ? [vs0] : val_left).each do |v0|					#V0, v1, v2, v3 are the values for the 4 variables
			((vs1) ? [vs1] : val_left - [v0]).each do |v1|
				((vs2) ? [vs2] : val_left - [v0, v1]).each do |v2|
					((vs3) ? [vs3] : val_left - [v0, v1, v2]).each do |v3|
						if proc.call(v0, v1, v2, v3)	#Testing the equation
							sol[j0], sol[j1], sol[j2], sol[j3] = v0, v1, v2, v3 
							lst_newsol.push sol.clone
						end	
						@nb_try += 1
					end
				end	
			end
		end	
	end
	@lst_sol = lst_newsol		#New active set of solutions
end

#Integer division - if numbers don't divide exactly, then make sure the equation test fails
def self.int_div(v1, v2) ; r = v1 / v2 ; (r * v2 == v1) ? r : 100000 ; end

#----------------------------------------------------------------------------------------
# Menu and load (called once)
#----------------------------------------------------------------------------------------

unless $PuzzleJimSolo_byFredo6____loaded
	UI.menu("Plugins").add_item("Puzzle Jim - Solo (Fredo6)") { startup }
	$PuzzleJimSolo_byFredo6____loaded = true
end

end #module PuzzleJimSolo_byFredo6